# demo

#  pip install mawiisurv
import numpy as np
import mawiisurv
import torch

# function input
'''
Main input variables:
    X: (n,p) array, covariates
    Z: (n,m) array, instrumental variables
    A: (n,) array, treatment
    Y: (n,) array, outcome
    censor_delta: (n,) array, censor indicator, 1 for uncensor and 0 for censor
    h: window for local KM estimation, and default setting is 1
    model_types: ['neural_network','linear_regression','random_forest','xgboost'], 4 choices of comparsion models, and default setting is ['neural_network']
    rho_function_names: ['ET','EL','CUE'], GEL functions, and default setting is ['EL']

Other DNN setting: 
    hidden_layers=[50, 50],
    learning_rate=0.0005,
    weight_decay=0.0001,
    batch_size=256,
    dropout_rate=0,
    patience=5,
    epochs=1000,
    validation_split=0.05,
    shuffle=False,
    device='cpu'
'''

# two main functions:
'''
'mawii_noncensor' for complete data
'mawii_censor' for right censor data
'''

# function output
'''
function will return a list of result, including 
    'beta' : estimation of treatment effect
    'se' : estimated standard error
    'test' : overidentification test statistics
'''

# complete data
n = 10000
m = 20
p = 1
beta_0 = 0.4
device = 'cuda:0' if torch.cuda.is_available() else 'cpu'


X = np.random.uniform(0, 1, size=(n, p))
p_Z = [0.25, 0.5, 0.25]
Z = np.random.choice([0, 1, 2], size=(n, m), p=p_Z)

gamma = np.sqrt(0.2 / (1.5*m)) * np.random.normal(0, 1, size=m)
delta = np.sqrt(0.2 / (1.5*m)) * np.random.normal(0, 1, size=m)

epsilon_A = np.random.normal(0, 0.4, size=n)
epsilon_Y = np.random.normal(0, 0.4, size=n)
U = np.random.normal(0, 0.6, size=n)

A = Z @ gamma + U + (1 + Z @ delta)*epsilon_A
Y = beta_0*A + np.sum(X,axis=1) - U  + epsilon_Y

result_noncensor = mawiisurv.mawii_noncensor(X, Z, A, Y, 
                          model_types= ['neural_network'], # more options: ['neural_network','linear_regression','random_forest','xgboost'],
                          rho_function_names=['ET'], # more options: ['ET','EL','CUE'],
                          device = device)

print(f"DNN+ET BETA: {result_noncensor['neural_network']['ET']['beta']:.3f}")
print(f"DNN+ET SE: {result_noncensor['neural_network']['ET']['se']:.3f}")
print(f"DNN+ET over-identification test: {result_noncensor['neural_network']['ET']['test']:.3f}")


# right-censor data simulation
T = beta_0*A + np.sum(X,axis=1) - U  + epsilon_Y
censor_rate = 0.4
rr = 0

while True:
    C = np.random.uniform(0+rr,5+rr, size=n)
    censor_delta = np.where(T <= C, 1, 0) 
    if np.mean(1-censor_delta) >=censor_rate+0.03:
        rr = rr + 0.1
    elif np.mean(1-censor_delta)<=censor_rate-0.03:
        rr = rr - 0.1
    else:
         break
        
Y = np.minimum(T,C)

result_censor = mawiisurv.mawii_censor(X, Z, A, Y, censor_delta, h = 1,
                                       model_types= ['neural_network'], # more options: ['neural_network','linear_regression','random_forest','xgboost'],
                                       rho_function_names=['ET'], # more options: ['ET','EL','CUE'],
                                       device = device)

print(f"DNN+ET BETA: {result_censor['neural_network']['ET']['beta']:.3f}")
print(f"DNN+ET SE: {result_censor['neural_network']['ET']['se']:.3f}")
print(f"DNN+ET over-identification test: {result_censor['neural_network']['ET']['test']:.3f}")





import numpy as np
from scipy.optimize import minimize
from sklearn.metrics.pairwise import pairwise_kernels
from numba import njit, prange
from .func import estimate_E_Z_given_X, estimate_h_functions, compute_beta_se, compute_beta_se_noncensor

def mawii_censor(
    X, Z, A, Y, censor_delta, h=1,
    model_types=['neural_network'],
    rho_function_names=['ET'],
    hidden_layers=[50, 50],
    learning_rate=0.0005,
    weight_decay=0.0001,
    batch_size=256,
    dropout_rate=0,
    patience=5,
    epochs=100,
    validation_split=0.05,
    shuffle=False,
    device='cpu',
):
    config_nn = {
        'hidden_layers': hidden_layers,
        'learning_rate': learning_rate,
        'weight_decay': weight_decay,
        'batch_size': batch_size,
        'dropout_rate': dropout_rate,
        'patience': patience,
        'epochs': epochs,
        'validation_split': validation_split,
        'shuffle': shuffle,
        'device': device
    }

    A = np.ravel(A)
    Y = np.ravel(Y)
    censor_delta = np.ravel(censor_delta)
    n = Z.shape[0]
    m = Z.shape[1]
    X_train, Z_train, A_train, Y_train, censor_train = X, Z, A, Y, censor_delta
    
    # Kernel & weight matrix
    OA_i = np.hstack((A_train[:,None], X_train, Z_train))
    K = pairwise_kernels(OA_i, OA_i, metric='rbf', gamma=2/h**2,n_jobs=-1)
    B = K/np.sum(K,axis=0)
    mask_y       = (Y_train[:,None] <= Y_train[None,:])
    mask_y_delta = mask_y & (censor_train[:,None] == 0)
    H = mask_y.astype(np.int8).dot(B)
    H = np.maximum(H, 0.01)
    
    @njit(parallel=True)
    def compute_Gn_fast(B, H, mask_y_delta):
        n = B.shape[0]
        Gn = np.ones(n, dtype=np.float64)
        for i in prange(n):
            g = 1.0
            for j in range(n):
                if mask_y_delta[j, i]:
                    g *= 1.0 - B[j, i] / H[j, i]
            Gn[i] = g if g >= 0.01 else 0.01
        return Gn
    
    Gn = compute_Gn_fast(B, H, mask_y_delta)
    IPW = censor_train/Gn  
    
    @njit(parallel=True)
    def compute_Phi_fast(Bji, Gn, Y_train, censor_train):
        m, n = Bji.shape
    
        Hji = np.empty((m, n), dtype=np.float64)
        for i in prange(m):
            yi = Y_train[i]
            for j in range(n):
                acc = 0.0
                for l in range(m):
                    if yi <= Y_train[l]:
                        acc += Bji[l, j]
                Hji[i, j] = acc if acc >= 0.1 else 0.1
    
        Phi = np.zeros((m, n), dtype=np.float64)
        for i in prange(m):
            yi = Y_train[i]
            ci = censor_train[i]
            dm_i = 1.0 if ci == 0 else 0.0
    
            for j in range(n):
                mj = 1.0 if (yi <= Y_train[j] and dm_i == 1.0) else 0.0
                mj_over = mj / Hji[i, j]
    
                t1 = 0.0
                if dm_i == 1.0:
                    for k in range(m):
                        if Y_train[k] <= yi and Y_train[k] <= Y_train[j]:
                            hkj = Hji[k, j]
                            t1 += Bji[k, j] / (hkj * hkj)
    
                Phi[i, j] = Bji[i, j] * Gn[j] * (mj_over - t1)
    
        return Phi
    
    
    Phi = compute_Phi_fast(B, Gn, Y_train, censor_train)

    rho_funcs = {
        'ET': lambda v: -np.exp(v)+1,
        'EL': lambda v: np.log(1-v),
        'CUE': lambda v: -v - v**2/2
    }
    rho_derivatives = {
         'ET': {
             'rho': lambda v: np.sum(-np.exp(v) + 1),
             'rho_prime': lambda v: -np.exp(v),
             'rho_double_prime': lambda v: -np.exp(v),
         },
         'EL': {
             'rho': lambda v: np.sum(np.log(1 - v)),
             'rho_prime': lambda v: -1 / (1 - v),
             'rho_double_prime': lambda v: -1 / (1 - v) ** 2,
         },
         'CUE': {
             'rho': lambda v: np.sum(-v - v ** 2 / 2),
             'rho_prime': lambda v: -1 - v,
             'rho_double_prime': lambda v: -np.ones_like(v),
         },
     }
    
    results = {}
    
    for model in model_types:
        EZ = estimate_E_Z_given_X(X_train,Z_train,config_nn,model)
        RA,RY,h3,h4 = estimate_h_functions(X_train,Z_train,A_train,Y_train,config_nn,model)
        def compute_g(beta):
            b=beta[0]
            Delta=(RA*RY - b*RA**2)-(h3-b*h4)
            g=(Z_train - EZ)*Delta[:,None]
            g_d1=-(Z_train-EZ)*(RA**2-h4)[:,None]
            return g,g_d1
        results[model] = {}
        for rho_name in rho_function_names:
            def Q(beta):
                g,g_d1 = compute_g(beta)
                xi = B.T.dot(IPW[:,None]*g)
                psi = IPW[:,None]*(g-xi)+xi
                Sigma = psi.T.dot(psi)/psi.shape[0]
                Sigma_inv = np.linalg.pinv(Sigma+np.eye(Sigma.shape[0])*1e-3)
                lam = -psi.mean(0).dot(Sigma_inv)
                lam_psi=psi.dot(lam)
                if rho_name=='EL': lam_psi = np.minimum(0.9, lam_psi)
                Q_value = rho_funcs[rho_name](lam_psi).mean()
                return Q_value
    
            res = minimize(Q, np.random.rand(1), method='Powell', bounds=[(-10,10)])
            beta_hat=res.x[0]
            ghat,g_d1=compute_g(res.x)
            xi= B.T.dot(IPW[:,None]*ghat)
            xi_d1=B.T.dot(IPW[:,None]*g_d1)
            SE,VarQ= compute_beta_se(ghat,g_d1,Phi,IPW,Gn,xi,xi_d1,rho_derivatives,rho_name)
            tempQ = 2*n*Q(res.x)
            test=(tempQ - (m-1))/np.sqrt(2*(m-1)*(1+VarQ))
            results[model][rho_name]={'beta':beta_hat,'se':SE,'test':test}
    
    return results


def mawii_noncensor(X, Z, A, Y,
        model_types=['neural_network'],
        rho_function_names=['ET'],
        hidden_layers=[50, 50],
        learning_rate=0.0005,
        weight_decay=0.0001,
        batch_size=256,
        dropout_rate=0,
        patience=5,
        epochs=100,
        validation_split=0.05,
        shuffle=False,
        device='cpu',
        ):
    
    config_nn = {
        'hidden_layers': hidden_layers,
        'learning_rate': learning_rate,
        'weight_decay': weight_decay,
        'batch_size': batch_size,
        'dropout_rate': dropout_rate,
        'patience': patience,
        'epochs': epochs,
        'validation_split': validation_split,
        'shuffle': shuffle,
        'device': device
    }
    A = np.ravel(A)
    Y = np.ravel(Y)
    n = Z.shape[0]
    m = Z.shape[1]
    X_train, Z_train, A_train, Y_train = X, Z, A, Y
    
    rho_funcs = {
        'ET': lambda v: -np.exp(v)+1,
        'EL': lambda v: np.log(1-v),
        'CUE': lambda v: -v - v**2/2
    }
    rho_derivatives = {
         'ET': {
             'rho': lambda v: np.sum(-np.exp(v) + 1),
             'rho_prime': lambda v: -np.exp(v),
             'rho_double_prime': lambda v: -np.exp(v),
         },
         'EL': {
             'rho': lambda v: np.sum(np.log(1 - v)),
             'rho_prime': lambda v: -1 / (1 - v),
             'rho_double_prime': lambda v: -1 / (1 - v) ** 2,
         },
         'CUE': {
             'rho': lambda v: np.sum(-v - v ** 2 / 2),
             'rho_prime': lambda v: -1 - v,
             'rho_double_prime': lambda v: -np.ones_like(v),
         },
     }
    
    results = {}
    
    
    for model in model_types:
        EZ = estimate_E_Z_given_X(X_train,Z_train,config_nn,model)
        RA,RY,h3,h4 = estimate_h_functions(X_train,Z_train,A_train,Y_train,config_nn,model)
        def compute_g(beta):
            b=beta[0]
            Delta=(RA*RY - b*RA**2)-(h3-b*h4)
            g=(Z_train - EZ)*Delta[:,None]
            g_d1=-(Z_train-EZ)*(RA**2-h4)[:,None]
            return g,g_d1
        results[model] = {}
        for rho_name in rho_function_names:
            def Q(b):
                g,g_d1 = compute_g(b)
                Omega = g.T.dot(g)/g.shape[0]
                inv = np.linalg.pinv(Omega+np.eye(Omega.shape[0])*1e-3)
                lam = -g.mean(0).dot(inv)
                lam_psi=g.dot(lam)
                if rho_name=='EL': lam_psi = np.minimum(0.99, lam_psi)
                return rho_funcs[rho_name](lam_psi).mean()
    
            res = minimize(Q, np.random.rand(1), method='Powell', bounds=[(-10,10)])
            beta_hat=res.x[0]
            ghat,g_d1=compute_g(res.x)
            SE= compute_beta_se_noncensor(ghat, g_d1, rho_derivatives, rho_name)
            tempQ = 2*n*Q(res.x)
            test=(tempQ - (m-1))/np.sqrt(2*(m-1))
            results[model][rho_name]={'beta':beta_hat,'se':SE,'test':test}
    
    return results




# func.py

import numpy as np
import torch
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from xgboost import XGBRegressor
from sklearn.multioutput import MultiOutputRegressor
from sklearn.metrics.pairwise import pairwise_kernels
from .dnn import train_regression_model 
import numpy as np


def estimate_E_Z_given_X(X_train, Z_train, config_regression_Z, model_type='neural_network'):
    m = Z_train.shape[1]
    E_Z_given_X_train = np.zeros((X_train.shape[0], m), dtype=float)

    if model_type == 'neural_network':
        model_ZX = train_regression_model(X_train, Z_train, config_regression_Z)
        with torch.no_grad():
            X_tensor_train = torch.tensor(X_train, dtype=torch.float32).to(config_regression_Z['device'])
            E_Z_given_X_train = model_ZX(X_tensor_train).cpu().numpy()
            
    elif model_type == 'random_forest':
        # Use RandomForestRegressor with multi-output support
        model_ZX = RandomForestRegressor(n_estimators=20, max_features=0.3,
    n_jobs=-1, max_depth=10, min_samples_split=5, random_state=0)
        model_ZX.fit(X_train, Z_train)
        E_Z_given_X_train = model_ZX.predict(X_train)
        
    elif model_type == 'xgboost':
        model_ZX = MultiOutputRegressor(XGBRegressor(n_estimators=20, random_state=0, verbosity=0))
        model_ZX.fit(X_train, Z_train)
        E_Z_given_X_train = model_ZX.predict(X_train)
        
    elif model_type == 'linear_regression':
        # Use LinearRegression with multi-output support
        model_ZX = LinearRegression()
        model_ZX.fit(X_train, Z_train)
        E_Z_given_X_train = model_ZX.predict(X_train)

    else:
        raise ValueError(f"Unknown model type: {model_type}")

    return  E_Z_given_X_train

def estimate_h_functions(X_train, Z_train, A_train, Y_train, config_regression_A, model_type='neural_network'):
    # Helper function to get the appropriate model
    def get_model():
        if model_type == 'neural_network':
            return train_regression_model
        elif model_type == 'random_forest':
            return RandomForestRegressor(n_estimators=100, max_features=0.3,
    n_jobs=-1, max_depth=10, min_samples_split=5, random_state=0)
        elif model_type == 'xgboost':
            return XGBRegressor(n_estimators=20, random_state=0, verbosity=0)
        elif model_type == 'linear_regression':
            return LinearRegression()
        elif model_type == 'mean':
            return None
        else:
            raise ValueError(f"Unknown model type: {model_type}")

    # Function to estimate h1
    def estimate_h1():
        XZ_input_train = np.hstack([X_train, Z_train])
        A_output_train = A_train.reshape(-1, 1)

        if model_type == 'neural_network':
            model_AXZ = train_regression_model(XZ_input_train, A_output_train, config_regression_A)
            with torch.no_grad():
                XZ_tensor_train = torch.tensor(XZ_input_train, dtype=torch.float32).to(config_regression_A['device'])
                h1_train = model_AXZ(XZ_tensor_train).cpu().numpy().flatten()
        elif model_type in ['random_forest', 'xgboost', 'linear_regression']:
            model_AXZ = get_model()
            model_AXZ.fit(XZ_input_train, A_output_train.ravel())
            h1_train = model_AXZ.predict(XZ_input_train).flatten()
        else :            
            h1_train = np.full(X_train.shape[0], A_train.mean())
        return h1_train

    # Function to estimate h2
    def estimate_h2():
        XZ_input_train = np.hstack([X_train, Z_train])
        Y_output_train = Y_train.reshape(-1, 1)

        if model_type == 'neural_network':
            model_YXZ = train_regression_model(XZ_input_train, Y_output_train, config_regression_A)
            with torch.no_grad():
                XZ_tensor_train = torch.tensor(XZ_input_train, dtype=torch.float32).to(config_regression_A['device'])
                h2_train = model_YXZ(XZ_tensor_train).cpu().numpy().flatten()
        elif model_type in ['random_forest', 'xgboost', 'linear_regression']:
            model_YXZ = get_model()
            model_YXZ.fit(XZ_input_train, Y_output_train.ravel())
            h2_train = model_YXZ.predict(XZ_input_train).flatten()
        else :            
            h2_train = np.full(X_train.shape[0], Y_train.mean())
        return  h2_train


    h1_train = estimate_h1()
    h2_train = estimate_h2()

    RA_train = A_train - h1_train
    RY_train = Y_train - h2_train

    # Function to estimate h3
    def estimate_h3():
        RA_RY_train = RA_train * RY_train
        RA_RY_output_train = RA_RY_train.reshape(-1, 1)

        if model_type == 'neural_network':
            model_RA_RY_X = train_regression_model(X_train, RA_RY_output_train, config_regression_A)
            with torch.no_grad():
                X_tensor_train = torch.tensor(X_train, dtype=torch.float32).to(config_regression_A['device'])
                h3_train = model_RA_RY_X(X_tensor_train).cpu().numpy().flatten()
        elif model_type in ['random_forest', 'xgboost', 'linear_regression']:
            model_RA_RY_X = get_model()
            model_RA_RY_X.fit(X_train, RA_RY_output_train.ravel())
            h3_train = model_RA_RY_X.predict(X_train).flatten()
        else:
            h3_train = np.full(X_train.shape[0], RA_RY_train.mean())
            
        return h3_train

    # Function to estimate h4
    def estimate_h4():
        RA_squared_train = RA_train ** 2
        RA_squared_output_train = RA_squared_train.reshape(-1, 1)

        if model_type == 'neural_network':
            model_RA_squared_X = train_regression_model(X_train, RA_squared_output_train, config_regression_A)
            with torch.no_grad():
                X_tensor_train = torch.tensor(X_train, dtype=torch.float32).to(config_regression_A['device'])
                h4_train = model_RA_squared_X(X_tensor_train).cpu().numpy().flatten()
        elif model_type in ['random_forest', 'xgboost', 'linear_regression']:
            model_RA_squared_X = get_model()
            model_RA_squared_X.fit(X_train, RA_squared_output_train.ravel())
            h4_train = model_RA_squared_X.predict(X_train).flatten()
        else:
            h4_train = np.full(X_train.shape[0], RA_squared_train.mean())
            
        return h4_train

    h3_train = estimate_h3()
    h4_train = estimate_h4()

    return RA_train, RY_train, h3_train, h4_train
    

def compute_beta_se(g_i, g_d1_i, Phin_ji, IPW, Gn_OA, xi_i, xi_d1_i, rho_derivatives, rho_name):
    
    g_bar = np.mean(g_i, axis=0)
    g_d1_bar = np.mean(g_d1_i, axis=0)
    xi_bar = np.mean(xi_i, axis=0)
    xi_d1_bar = np.mean(xi_d1_i, axis=0)
    psi_i = IPW[:,np.newaxis] * (g_i - xi_i) + xi_i
    psi_d1_i = IPW[:,np.newaxis] * (g_d1_i - xi_d1_i) + xi_d1_i
    psi_bar_i = np.mean(psi_i, axis=0)
    
    n_train, m = g_i.shape
    n_est = Phin_ji.shape[0]
    
    Omega = (1 / n_train) * (psi_i.T @ psi_i)  # m x m matrix
    Omega_inv = np.linalg.pinv(Omega+np.eye(m)*1e-3)

    Omega_d1 = (1 / n_train) * (psi_d1_i.T @ psi_i + psi_i.T @ psi_d1_i)  # m x m matrix
    Omega_d2 = (2 / n_train) * (psi_d1_i.T @ psi_d1_i)  # m x m matrix
    Omega_inv_d1 = -Omega_inv @ Omega_d1 @ Omega_inv
    Omega_inv_d2 = -2*Omega_inv_d1 @ Omega_d1 @ Omega_inv - Omega_inv @ Omega_d2 @ Omega_inv
    psi_bar = np.mean(psi_i, axis=0)
    psi_d1_bar = np.mean(psi_d1_i, axis=0)
    psi_bar_Omega_inv = -psi_bar @ Omega_inv  # m * 1
    psi_bar_Omega_inv_d1 = -psi_d1_bar @ Omega_inv - psi_bar @ Omega_inv_d1
    psi_bar_Omega_inv_d2 = -2*psi_d1_bar @ Omega_inv_d1 - psi_bar @ Omega_inv_d2
    
    psi_bar_dG = (g_bar-xi_bar)
    psi_d1_bar_dG = (g_d1_bar-xi_d1_bar)
    Omega_dG_inv = -Omega_inv@(np.outer(psi_bar_dG,psi_bar_i)+ np.outer(psi_bar_i,psi_bar_dG))@Omega_inv

    s_i = psi_i @ psi_bar_Omega_inv
    # Compute partial_s_i
    s_i_d1 = psi_d1_i @ psi_bar_Omega_inv + psi_i @ psi_bar_Omega_inv_d1  # n-dimensional vector
    s_i_d2 = 2* psi_d1_i @ psi_bar_Omega_inv_d1 + psi_i @ psi_bar_Omega_inv_d2  # n-dimensional vector
    # Get rho derivatives
    rho_prime = rho_derivatives[rho_name]['rho_prime']
    rho_double_prime = rho_derivatives[rho_name]['rho_double_prime']
    rho_prime_s_i = rho_prime(s_i)            # n-dimensional vector
    rho_double_prime_s_i = rho_double_prime(s_i)  # n-dimensional vector

    # Compute H (scalar)
    H = np.mean(rho_double_prime_s_i * (s_i_d1 ** 2) + rho_prime_s_i * s_i_d2)

    weights = rho_prime_s_i / np.sum(rho_prime_s_i)  # n-dimensional vector
    D = psi_d1_i.T @ weights  # m-dimensional vector
    # Compute variance of Gn-G
    term1 = rho_double_prime_s_i*(((xi_i-g_i)@ psi_bar_Omega_inv )* (psi_d1_i @ psi_bar_Omega_inv))*IPW/Gn_OA
    term2 = rho_prime_s_i*((xi_d1_i-g_d1_i)@ psi_bar_Omega_inv)*IPW /Gn_OA
    VarG =  (H ** -2) * np.sum((Phin_ji @ (term1+term2))** 2)/ n_train**2  +\
               2*(H ** -2) * np.sum((Phin_ji @ (term1+term2)) @ (psi_d1_i @ (-psi_bar_Omega_inv)) )/ n_train 

    Varbeta = (H ** -2) * (D.T @ Omega_inv @ D)/ n_train  # Scalar 
    SE = np.sqrt(Varbeta+VarG)  # Standard Error
    VarQ = 2*np.sum((Phin_ji @ (rho_prime_s_i*((xi_i-g_i)@ psi_bar_Omega_inv)*IPW /Gn_OA))** 2)/(n_train**2) 
    return SE, VarQ

def compute_beta_se_noncensor(g_i, G_i, rho_derivatives, rho_name):
    n_est, m = g_i.shape

    # Compute Omega_hat
    Omega = (1 / n_est) * (g_i.T @ g_i)  # m x m matrix
    Omega_inv = np.linalg.pinv(Omega)

    # Compute Omega_hat_1 and Omega_hat_2
    Omega_d1 = (1 / n_est) * (G_i.T @ g_i + g_i.T @ G_i)  # m x m matrix
    Omega_d2 = (2 / n_est) * (G_i.T @ G_i)  # m x m matrix
    Omega_inv_d1 = -Omega_inv @ Omega_d1 @ Omega_inv
    Omega_inv_d2 = -2*Omega_inv_d1 @ Omega_d1 @ Omega_inv - Omega_inv @ Omega_d2 @ Omega_inv
    g_bar = np.mean(g_i, axis=0)
    G_bar = np.mean(G_i, axis=0)
    g_bar_Omega_inv = -g_bar @ Omega_inv  # m * 1
    g_bar_Omega_inv_d1 = -G_bar @ Omega_inv - g_bar @ Omega_inv_d1
    g_bar_Omega_inv_d2 = -2*G_bar @ Omega_inv_d1 - g_bar @ Omega_inv_d2
    

    s_i = g_i @ g_bar_Omega_inv
    # Compute partial_s_i
    s_i_d1 = G_i @ g_bar_Omega_inv + g_i @ g_bar_Omega_inv_d1  # n-dimensional vector
    s_i_d2 = 2* G_i @ g_bar_Omega_inv_d1 + g_i @ g_bar_Omega_inv_d2  # n-dimensional vector
    # Get rho derivatives
    rho_prime = rho_derivatives[rho_name]['rho_prime']
    rho_double_prime = rho_derivatives[rho_name]['rho_double_prime']

    # Compute rho'(s_i) and rho''(s_i)
    rho_prime_s_i = rho_prime(s_i)            # n-dimensional vector
    rho_double_prime_s_i = rho_double_prime(s_i)  # n-dimensional vector

    # Compute H (scalar)
    H = np.mean(rho_double_prime_s_i * (s_i_d1 ** 2) + rho_prime_s_i * s_i_d2)

    # Compute weights w_i
    weights = rho_prime_s_i / np.sum(rho_prime_s_i)  # n-dimensional vector

    # Compute D_0 (m-dimensional vector)
    D = G_i.T @ weights  # m-dimensional vector

    # Compute V/n (scalar)
    V = (H ** -2) * D.T @ Omega_inv @ D  # Scalar
    SE = np.sqrt(V / n_est)  # Standard Error
    

    return SE


# dnn.py

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset, random_split
import numpy as np

# Neural network model
class RegressionNeuralNetwork(nn.Module):
    def __init__(self, input_size, hidden_layers, output_size, dropout_rate=0.0):
        super(RegressionNeuralNetwork, self).__init__()
        layers = []
        prev_size = input_size

        for h in hidden_layers:
            linear_layer = nn.Linear(prev_size, h)
            layers.append(linear_layer)
            layers.append(nn.ReLU())
            if dropout_rate > 0:
                layers.append(nn.Dropout(dropout_rate))
            prev_size = h

        output_layer = nn.Linear(prev_size, output_size)
        layers.append(output_layer)

        self.network = nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)

# Training function with multiple initializations
def train_regression_model(X, Y, config, num_initializations=1):
    # Dynamically get input_size and output_size
    input_size = X.shape[1]  # Number of input features
    output_size = Y.shape[1]  # Output size

    # Convert to Tensor
    X_tensor = torch.tensor(X, dtype=torch.float32)
    Y_tensor = torch.tensor(Y, dtype=torch.float32)

    # Move tensors to device only when creating DataLoader to avoid CUDA memory issues
    dataset = TensorDataset(X_tensor, Y_tensor)

    # Split dataset into training and validation sets
    val_size = int(len(dataset) * config['validation_split'])
    train_size = len(dataset) - val_size
    # Ensure reproducibility in dataset splitting
    train_dataset, val_dataset = random_split(dataset, [train_size, val_size],
                                              generator=torch.Generator().manual_seed(config.get('seed', 0)))

    best_model_state = None
    best_val_loss_overall = float('inf')
    best_model_init = None

    for init in range(num_initializations):
        # Set different seeds for different initializations
        seed = config.get('seed', 0) + init  # Modify seed for each initialization
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)
        # Define the model
        model = RegressionNeuralNetwork(input_size, config['hidden_layers'], output_size, config['dropout_rate']).to(config['device'])

        # Define loss function and optimizer
        criterion = nn.MSELoss(reduction='mean')
        optimizer = optim.Adam(model.parameters(), lr=config['learning_rate'], weight_decay=config['weight_decay'])

        # Create DataLoader for training and validation (move data to device here)
        train_loader = DataLoader(train_dataset, batch_size=config['batch_size'], shuffle=config['shuffle'])
        val_loader = DataLoader(val_dataset, batch_size=config['batch_size'])

        best_val_loss = float('inf')
        patience_counter = 0
        best_model_state_init = None

        # Training loop
        for epoch in range(config['epochs']):
            model.train()
            total_loss = 0

            # Training step
            for X_batch, Y_batch in train_loader:
                X_batch = X_batch.to(config['device'])
                Y_batch = Y_batch.to(config['device'])
                optimizer.zero_grad()
                outputs = model(X_batch)
                loss = criterion(outputs, Y_batch)
                loss.backward()
                optimizer.step()
                total_loss += loss.item()

            # Validation step
            model.eval()
            val_loss = 0
            with torch.no_grad():
                for X_val, Y_val in val_loader:
                    X_val = X_val.to(config['device'])
                    Y_val = Y_val.to(config['device'])
                    outputs = model(X_val)
                    loss = criterion(outputs, Y_val)
                    val_loss += loss.item()

            val_loss /= len(val_loader)
            # print(f'Initialization {init+1}, Epoch {epoch+1}, Validation Loss: {val_loss:.6f}, Training Loss: {total_loss/len(train_loader):.6f}')

            # Early stopping mechanism
            if val_loss < best_val_loss:
                best_val_loss = val_loss
                patience_counter = 0

                # Save the best model state for this initialization
                best_model_state_init = {
                    'epoch': epoch + 1,
                    'model_state_dict': model.state_dict(),
                    'optimizer_state_dict': optimizer.state_dict(),
                    'val_loss': best_val_loss,
                }
            else:
                patience_counter += 1

            if patience_counter >= config['patience']:
                # print(f"Early stopping at epoch {epoch+1}")
                break

        # After training for this initialization, check if this is the best overall
        if best_val_loss < best_val_loss_overall:
            best_val_loss_overall = best_val_loss
            best_model_state = best_model_state_init
            best_model_init = init

    # After all initializations, load the best model state into a new model
    if best_model_state is not None:
        # Re-initialize the model with the seed corresponding to the best initialization
        seed = config.get('seed', 0) + best_model_init
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)
        best_model = RegressionNeuralNetwork(input_size, config['hidden_layers'], output_size, config['dropout_rate']).to(config['device'])
        best_model.load_state_dict(best_model_state['model_state_dict'])
        # Optionally, you can also return the optimizer state if needed
        # optimizer.load_state_dict(best_model_state['optimizer_state_dict'])
        # print(f"Loaded best model from initialization {best_model_init+1} with validation loss {best_val_loss_overall:.6f}")
    else:
        print("No valid model was found.")
        return None

    return best_model

    
# Classification Neural Network
class ClassificationNeuralNetwork(nn.Module):
    def __init__(self, input_size, hidden_layers, num_classes, dropout_rate=0.0):
        super(ClassificationNeuralNetwork, self).__init__()
        layers = []
        prev_size = input_size

        # Add hidden layers
        for h in hidden_layers:
            layers.append(nn.Linear(prev_size, h))
            layers.append(nn.ReLU())
            if dropout_rate > 0:
                layers.append(nn.Dropout(dropout_rate))
            prev_size = h

        # Add output layer
        layers.append(nn.Linear(prev_size, num_classes))

        self.network = nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)



# Training function for classification
def train_classification_model(X, Y, config):
    # Dynamically get input_size and num_classes
    input_size = X.shape[1]
    # Convert Y to tensor before calling torch.unique
    Y_tensor = torch.tensor(Y, dtype=torch.long).to(config['device']).view(-1)  # For classification, Y should be long tensor

    num_classes = len(torch.unique(Y_tensor))

    # Create dataset
    dataset = TensorDataset(torch.tensor(X, dtype=torch.float32).to(config['device']), Y_tensor)

    # Split dataset into training and validation sets
    val_size = int(len(dataset) * config['validation_split'])
    train_size = len(dataset) - val_size
    train_dataset, val_dataset = random_split(dataset, [train_size, val_size])

    # Create DataLoader for training and validation
    train_loader = DataLoader(train_dataset, batch_size=config['batch_size'], shuffle=config['shuffle'])
    val_loader = DataLoader(val_dataset, batch_size=config['batch_size'])

    # Define the model
    model = ClassificationNeuralNetwork(input_size, config['hidden_layers'], num_classes, config['dropout_rate']).to(config['device'])

    # Define loss function and optimizer
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=config['learning_rate'], weight_decay=config['weight_decay'])

    best_val_loss = float('inf')
    patience_counter = 0

    # Training loop
    for epoch in range(config['epochs']):
        model.train()
        total_loss = 0

        # Training step
        for X_batch, Y_batch in train_loader:
            optimizer.zero_grad()
            outputs = model(X_batch)
            loss = criterion(outputs, Y_batch)
            loss.backward()
            optimizer.step()
            total_loss += loss.item()

        # Validation step
        model.eval()
        val_loss = 0
        with torch.no_grad():
            for X_val, Y_val in val_loader:
                outputs = model(X_val)
                loss = criterion(outputs, Y_val)
                val_loss += loss.item()

        val_loss /= len(val_loader)
        # print(val_loss)
        # Early stopping mechanism
        if val_loss < best_val_loss:
            best_val_loss = val_loss
            patience_counter = 0
        else:
            patience_counter += 1

        if patience_counter >= config['patience']:
            break

    return model



"""
genius_censor package
"""

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version
__version__ = version(__package__ or "mawiisurv")

from .mawii_surv import mawii_censor,mawii_noncensor

__all__ = ["mawii_censor", "mawii_noncensor"]